"""Position reconstruction for Xenon-nT"""

import os
import tempfile
import tarfile
import numpy as np
import strax
import straxen
from warnings import warn
export, __all__ = strax.exporter()

DEFAULT_POSREC_ALGO_OPTION = tuple([strax.Option("default_reconstruction_algorithm",
                 help="default reconstruction algorithm that provides (x,y)",
                 default="mlp",
                 )])

@export
@strax.takes_config(
    strax.Option('min_reconstruction_area',
                 help='Skip reconstruction if area (PE) is less than this',
                 default=10),
    strax.Option('n_top_pmts', default=straxen.n_top_pmts,
                 help="Number of top PMTs")
)

class PeakPositionsBaseNT(strax.Plugin):
    """
    Base class for reconstructions.
    This class should only be used when subclassed for the different
    algorithms. Provides x_algorithm, y_algorithm for all peaks > than
    min-reconstruction area based on the top array.
    """
    depends_on = ('peaks',)
    algorithm = None
    compressor = 'zstd'
    # Using parallel = 'process' is not allowed as we cannot Pickle
    # self.model during multiprocessing (to fix?)
    parallel = True
    __version__ = '0.0.0'

    def infer_dtype(self):
        if self.algorithm is None:
            raise NotImplementedError(f'Base class should not be used without '
                                      f'algorithm as done in {__class__.__name__}')
        dtype = [('x_' + self.algorithm, np.float32,
                  f'Reconstructed {self.algorithm} S2 X position (cm), uncorrected'),
                 ('y_' + self.algorithm, np.float32,
                  f'Reconstructed {self.algorithm} S2 Y position (cm), uncorrected')]
        dtype += strax.time_fields
        return dtype

    def setup(self):
        self.model_file = self._get_model_file_name()
        if self.model_file is None:
            warn(f'No file provided for {self.algorithm}. Setting all values '
                 f'for {self.provides} to None.')
            # No further setup required
            return

        # Load the tensorflow model
        import tensorflow as tf
        if os.path.exists(self.model_file):
            print(f"Path is local. Loading {self.algorithm} TF model locally "
                  f"from disk.")
        else:
            downloader = straxen.MongoDownloader()
            try:
                self.model_file = downloader.download_single(self.model_file)
            except straxen.mongo_storage.CouldNotLoadError as e:
                raise RuntimeError(f'Model files {self.model_file} is not found') from e
        with tempfile.TemporaryDirectory() as tmpdirname:
            tar = tarfile.open(self.model_file, mode="r:gz")
            tar.extractall(path=tmpdirname)
            self.model = tf.keras.models.load_model(tmpdirname)

    def compute(self, peaks):
        result = np.ones(len(peaks), dtype=self.dtype)
        result['time'], result['endtime'] = peaks['time'], strax.endtime(peaks)

        result['x_' + self.algorithm] *= float('nan')
        result['y_' + self.algorithm] *= float('nan')

        if self.model_file is None:
            # This plugin is disabled since no model is provided
            return result

        # Keep large peaks only
        peak_mask = peaks['area'] > self.config['min_reconstruction_area']
        if not np.sum(peak_mask):
            # Nothing to do, and .predict crashes on empty arrays
            return result

        # Getting actual position reconstruction
        _in = peaks['area_per_channel'][peak_mask, 0:self.config['n_top_pmts']]
        with np.errstate(divide='ignore', invalid='ignore'):
            _in = _in / np.max(_in, axis=1).reshape(-1, 1)
        _in = _in.reshape(-1, self.config['n_top_pmts'])
        _out = self.model.predict(_in)

        # writing output to the result
        result['x_' + self.algorithm][peak_mask] = _out[:, 0]
        result['y_' + self.algorithm][peak_mask] = _out[:, 1]
        return result

    def _get_model_file_name(self):

        config_file = f'{self.algorithm}_model'
        model_from_config = self.config.get(config_file, 'No file')
        if model_from_config == 'No file':
            raise ValueError(f'{__class__.__name__} should have {config_file} '
                             f'provided as an option.')
        if isinstance(model_from_config, str) and os.path.exists(model_from_config):
            # Allow direct path specification
            return model_from_config
        if model_from_config is None:
            # Allow None to be specified (disables processing for given posrec)
            return model_from_config

        # Use CMT
        model_file = straxen.get_correction_from_cmt(self.run_id, model_from_config)
        return model_file

@export
@strax.takes_config(
    strax.Option('mlp_model',
                 help='Neural network model.' 
                      'If CMT, specify as (mlp_model, ONLINE, True)'
                      'Set to None to skip the computation of this plugin.',
                 default=('mlp_model', "ONLINE", True)
                )
)
class PeakPositionsMLP(PeakPositionsBaseNT):
    """Multilayer Perceptron (MLP) neural net for position reconstruction"""
    provides = "peak_positions_mlp"
    algorithm = "mlp"


@export
@strax.takes_config(
    strax.Option('gcn_model',
                 help='Neural network model.' 
                      'If CMT, specify as  (gcn_model, ONLINE, True)'
                      'Set to None to skip the computation of this plugin.',
                 default=('gcn_model', "ONLINE", True)
                )
)
class PeakPositionsGCN(PeakPositionsBaseNT):
    """Graph Convolutional Network (GCN) neural net for position reconstruction"""
    provides = "peak_positions_gcn"
    algorithm = "gcn"
    __version__ = '0.0.1'


@export
@strax.takes_config(
    strax.Option('cnn_model',
                 help='Neural network model.' 
                      'If CMT, specify as (cnn_model, ONLINE, True)'
                      'Set to None to skip the computation of this plugin.',
                 default=('cnn_model', "ONLINE", True)
                )
)
class PeakPositionsCNN(PeakPositionsBaseNT):
    """Convolutional Neural Network (CNN) neural net for position reconstruction"""
    provides = "peak_positions_cnn"
    algorithm = "cnn"
    __version__ = '0.0.1'


@export
@strax.takes_config(
    *DEFAULT_POSREC_ALGO_OPTION
)
class PeakPositionsNT(strax.MergeOnlyPlugin):
    """
    Merge the reconstructed algorithms of the different algorithms 
    into a single one that can be used in Event Basics.
    
    Select one of the plugins to provide the 'x' and 'y' to be used 
    further down the chain. Since we already have the information
    needed here, there is no need to wait until events to make the
    decision.
    
    Since the computation is trivial as it only combined the three 
    input plugins, don't save this plugins output.
    """
    provides = "peak_positions"
    depends_on = ("peak_positions_cnn", "peak_positions_mlp", "peak_positions_gcn")
    save_when = strax.SaveWhen.NEVER
    __version__ = '0.0.0'

    def infer_dtype(self):
        dtype = strax.merged_dtype([self.deps[d].dtype_for(d) for d in self.depends_on])
        dtype += [('x', np.float32, 'Reconstructed S2 X position (cm), uncorrected'),
                  ('y', np.float32, 'Reconstructed S2 Y position (cm), uncorrected')]
        return dtype

    def compute(self, peaks):
        result = {dtype: peaks[dtype] for dtype in peaks.dtype.names}
        algorithm = self.config['default_reconstruction_algorithm']
        if not 'x_' + algorithm in peaks.dtype.names:
            raise ValueError
        for xy in ('x', 'y'):
            result[xy] = peaks[f'{xy}_{algorithm}']
        return result



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import strax
import straxen
from datetime import datetime
import pytz

export, __all__ = strax.exporter()

# Default attributes to display in the event_display (looks little
# complicated but just repeats same fields for S1 S1)
# Should be of form as below where {v} wil be filled with the value of
# event['key']:
#  (('key', '{v} UNIT'), ..)
PEAK_DISPLAY_DEFAULT_INFO = sum([[(k.format(i=s_i), u) for k, u in
                                  (('cs{i}', '{v:.2f} PE'),
                                   ('s{i}_area', '{v:.2f} PE'),
                                   ('alt_cs{i}', '{v:.2f} PE'),
                                   ('s{i}_n_channels', '{v}'),
                                   ('s{i}_area_fraction_top', '{v:.2f}'),
                                   ('s{i}_range_50p_area', '{v:.1f}'),
                                   )] for s_i in (1, 2)], [])
EVENT_DISPLAY_DEFAULT_INFO = (('time', '{v} ns'),
                              ('endtime', '{v} ns'),
                              ('event_number', '{v}'),
                              ('x', '{v:.2f} cm'),
                              ('y', '{v:.2f} cm'),
                              ('z', '{v:.2f} cm'),
                              ('r', '{v:.2f} cm'),
                              ('theta', '{v:.2f} rad'),
                              ('drift_time', '{v} ns'),
                              ('alt_s1_interaction_drift_time', '{v} ns'),
                              ('alt_s2_interaction_drift_time', '{v} ns')
                              )


# Don't be smart with the arguments, since it is a minianalyses we
# need to have all the arguments
@straxen.mini_analysis(requires=('event_info',))
def event_display_simple(context,
                         run_id,
                         events,
                         to_pe,
                         records_matrix=True,
                         s2_fuzz=50,
                         s1_fuzz=0,
                         max_peaks=500,
                         xenon1t=False,
                         display_peak_info=PEAK_DISPLAY_DEFAULT_INFO,
                         display_event_info=EVENT_DISPLAY_DEFAULT_INFO,
                         s1_hp_kwargs=None,
                         s2_hp_kwargs=None,
                         event_time_limit=None,
                         plot_all_positions=True,
                         ):
    """
    {event_docs}
    {event_returns}
    """
    fig = plt.figure(figsize=(12, 8), facecolor="white")
    grid = plt.GridSpec(2, 3, hspace=0.5)
    axes = dict()
    axes["ax_s1"] = fig.add_subplot(grid[0, 0])
    axes["ax_s2"] = fig.add_subplot(grid[0, 1])
    axes["ax_s2_hp_t"] = fig.add_subplot(grid[0, 2])
    axes["ax_ev"] = fig.add_subplot(grid[1, :])

    return _event_display(context,
                          run_id,
                          events,
                          to_pe,
                          axes=axes,
                          records_matrix=records_matrix,
                          s2_fuzz=s2_fuzz,
                          s1_fuzz=s1_fuzz,
                          max_peaks=max_peaks,
                          xenon1t=xenon1t,
                          display_peak_info=display_peak_info,
                          display_event_info=display_event_info,
                          s1_hp_kwargs=s1_hp_kwargs,
                          s2_hp_kwargs=s2_hp_kwargs,
                          event_time_limit=event_time_limit,
                          plot_all_positions=plot_all_positions,
                          )


# Don't be smart with the arguments, since it is a minianalyses we
# need to have all the arguments
@straxen.mini_analysis(requires=('event_info',))
def event_display(context,
                  run_id,
                  events,
                  to_pe,
                  records_matrix=True,
                  s2_fuzz=50,
                  s1_fuzz=0,
                  max_peaks=500,
                  xenon1t=False,
                  display_peak_info=PEAK_DISPLAY_DEFAULT_INFO,
                  display_event_info=EVENT_DISPLAY_DEFAULT_INFO,
                  s1_hp_kwargs=None,
                  s2_hp_kwargs=None,
                  event_time_limit=None,
                  plot_all_positions=True,
                  ):
    """
    {event_docs}
    {event_returns}
    """
    if records_matrix not in ('raw', True, False):
        raise ValueError('Choose either "raw", True or False for records_matrix')
    if ((records_matrix == 'raw' and not context.is_stored(run_id, 'raw_records')) or
            (isinstance(records_matrix, bool) and not context.is_stored(run_id,
                                                                        'records'))):  # noqa
        print("(raw)records not stored! Not showing records_matrix")
        records_matrix = False
    # Convert string to int to allow plots to be enlarged for extra panel
    _rr_resize_int = int(bool(records_matrix))

    fig = plt.figure(figsize=(25, 21 if _rr_resize_int else 16),
                     facecolor='white')
    grid = plt.GridSpec((2 + _rr_resize_int), 1, hspace=0.1 + 0.1 * _rr_resize_int,
                        height_ratios=[1.5, 0.5, 0.5][:2 + _rr_resize_int]
                        )

    # S1, S2, hitpatterns
    gss_0 = gridspec.GridSpecFromSubplotSpec(2, 4, subplot_spec=grid[0], wspace=0.25, hspace=0.4)
    ax_s1 = fig.add_subplot(gss_0[0])
    ax_s2 = fig.add_subplot(gss_0[1])
    ax_s1_hp_t = fig.add_subplot(gss_0[2])
    ax_s1_hp_b = fig.add_subplot(gss_0[3])
    ax_s2_hp_t = fig.add_subplot(gss_0[6])
    ax_s2_hp_b = fig.add_subplot(gss_0[7])

    # Peak & event info
    ax_event_info = fig.add_subplot(gss_0[4])
    ax_peak_info = fig.add_subplot(gss_0[5])

    # All peaks in event
    gss_1 = gridspec.GridSpecFromSubplotSpec(1, 1, subplot_spec=grid[1])
    ax_ev = fig.add_subplot(gss_1[0])
    ax_rec = None

    # (raw)records matrix (optional)
    if records_matrix and ax_rec is not None:
        gss_2 = gridspec.GridSpecFromSubplotSpec(1, 1, subplot_spec=grid[2])
        ax_rec = fig.add_subplot(gss_2[0])
    axes = dict(
        ax_s1=ax_s1,
        ax_s2=ax_s2,
        ax_s1_hp_t=ax_s1_hp_t,
        ax_s1_hp_b=ax_s1_hp_b,
        ax_event_info=ax_event_info,
        ax_peak_info=ax_peak_info,
        ax_s2_hp_t=ax_s2_hp_t,
        ax_s2_hp_b=ax_s2_hp_b,
        ax_ev=ax_ev,
        ax_rec=ax_rec)

    return _event_display(context,
                          run_id,
                          events,
                          to_pe,
                          axes=axes,
                          records_matrix=records_matrix,
                          s2_fuzz=s2_fuzz,
                          s1_fuzz=s1_fuzz,
                          max_peaks=max_peaks,
                          xenon1t=xenon1t,
                          display_peak_info=display_peak_info,
                          display_event_info=display_event_info,
                          s1_hp_kwargs=s1_hp_kwargs,
                          s2_hp_kwargs=s2_hp_kwargs,
                          event_time_limit=event_time_limit,
                          plot_all_positions=plot_all_positions,
                          )


def _event_display(context,
                   run_id,
                   events,
                   to_pe,
                   axes=None,
                   records_matrix=True,
                   s2_fuzz=50,
                   s1_fuzz=0,
                   max_peaks=500,
                   xenon1t=False,
                   display_peak_info=PEAK_DISPLAY_DEFAULT_INFO,
                   display_event_info=EVENT_DISPLAY_DEFAULT_INFO,
                   s1_hp_kwargs=None,
                   s2_hp_kwargs=None,
                   event_time_limit=None,
                   plot_all_positions=True,
                   ):
    """{event_docs}
    :param axes: if a dict of matplotlib axes (w/ same keys as below,
        and empty/None for panels not filled)
    {event_returns} 
    """
    if len(events) != 1:
        raise ValueError(f'Found {len(events)} only request one')
    event = events[0]

    if not context.is_stored(run_id, 'peaklets'):
        raise strax.DataNotAvailable(f'peaklets not available for {run_id}')

    if axes is None:
        raise ValueError(f'No axes provided')
    ax_s1 = axes.get("ax_s1", None)
    ax_s2 = axes.get("ax_s2", None)
    ax_s1_hp_t = axes.get("ax_s1_hp_t", None)
    ax_s1_hp_b = axes.get("ax_s1_hp_b", None)
    ax_s2_hp_t = axes.get("ax_s2_hp_t", None)
    ax_s2_hp_b = axes.get("ax_s2_hp_b", None)
    ax_event_info = axes.get("ax_event_info", None)
    ax_peak_info = axes.get("ax_peak_info", None)
    ax_ev = axes.get("ax_ev", None)
    ax_rec = axes.get("ax_rec", None)

    # titles
    for ax, title in zip([ax_s1, ax_s1_hp_t, ax_s1_hp_b,
                          ax_s2, ax_s2_hp_t, ax_s2_hp_b,
                          ax_event_info, ax_peak_info],
                         ["Main S1", "S1 top", "S1 bottom",
                          "Main S2", "S2 top", "S2 bottom",
                          "Event info", "Peak info"]):
        if ax is not None:
            ax.set_title(title)

    # Parse the hit pattern options
    # Convert to dict (not at function definition because of mutable defaults)
    if s1_hp_kwargs is None:
        s1_hp_kwargs = {}
    if s2_hp_kwargs is None:
        s2_hp_kwargs = {}

    # Hit patterns options:
    for hp_opt, color_map in ((s1_hp_kwargs, "Blues"), (s2_hp_kwargs, "Greens")):
        _common_opt = dict(xenon1t=xenon1t,
                           pmt_label_color='lightgrey',
                           log_scale=True,
                           vmin=0.1,
                           s=(250 if records_matrix else 220),
                           pmt_label_size=7,
                           edgecolor='grey',
                           dead_pmts=np.argwhere(to_pe == 0),
                           cmap=color_map)
        # update s1 & S2 hit pattern kwargs with _common_opt if not
        # specified by the user
        for k, v in _common_opt.items():
            if k not in hp_opt:
                hp_opt[k] = v

    # S1
    if events['s1_area'] != 0:
        if ax_s1 is not None:
            plt.sca(ax_s1)
            context.plot_peaks(run_id,
                               time_range=(events['s1_time'] - s1_fuzz,
                                           events['s1_endtime'] + s1_fuzz),
                               single_figure=False)

        # Hit pattern plots
        area = context.get_array(run_id, 'peaklets',
                                 time_range=(events['s1_time'],
                                             events['s1_endtime']),
                                 keep_columns=('area_per_channel', 'time', 'dt', 'length'),
                                 progress_bar=False,
                                 )
        for ax, array in ((ax_s1_hp_t, 'top'), (ax_s1_hp_b, 'bottom')):
            if ax is not None:
                plt.sca(ax)
                straxen.plot_on_single_pmt_array(c=np.sum(area['area_per_channel'], axis=0),
                                                 array_name=array,
                                                 **s1_hp_kwargs)
                # Mark reconstructed position
                plt.scatter(event['x'], event['y'], marker='X', s=100, c='k')

    # S2
    if event['s2_area'] != 0:
        if ax_s2 is not None:
            plt.sca(ax_s2)
            context.plot_peaks(run_id,
                               time_range=(events['s2_time'] - s2_fuzz,
                                           events['s2_endtime'] + s2_fuzz),
                               single_figure=False)

        # Hit pattern plots
        area = context.get_array(run_id, 'peaklets',
                                 time_range=(events['s2_time'],
                                             events['s2_endtime']),
                                 keep_columns=('area_per_channel', 'time', 'dt', 'length'),
                                 progress_bar=False,
                                 )
        for axi, (ax, array) in enumerate([(ax_s2_hp_t, 'top'), (ax_s2_hp_b, 'bottom')]):
            if ax is not None:
                plt.sca(ax)
                straxen.plot_on_single_pmt_array(c=np.sum(area['area_per_channel'], axis=0),
                                                 array_name=array,
                                                 **s2_hp_kwargs)
                # Mark reconstructed position (corrected)
                plt.scatter(event['x'], event['y'], marker='X', s=100, c='k')
                if not xenon1t and axi == 0 and plot_all_positions:
                    _scatter_rec(event)

    # Fill panels with peak/event info
    for it, (ax, labels_and_unit) in enumerate([(ax_event_info, display_event_info),
                                                (ax_peak_info, display_peak_info)]):
        if ax is not None:
            for i, (_lab, _unit) in enumerate(labels_and_unit):
                coord = 0.01, 0.9 - 0.9 * i / len(labels_and_unit)
                ax.text(*coord, _lab[:24], va='top', zorder=-10)
                ax.text(coord[0] + 0.5, coord[1],
                        _unit.format(v=event[_lab]), va='top', zorder=-10)
                # Remove axes and labels from panel
                ax.set_xticks([])
                ax.set_yticks([])
                _ = [s.set_visible(False) for s in ax.spines.values()]

    # Plot peaks in event
    ev_range = None
    if ax_ev is not None:
        plt.sca(ax_ev)
        if event_time_limit is None:
            time_range = (events['time'], events['endtime'])
        else:
            time_range = event_time_limit

        context.plot_peaks(run_id,
                           time_range=time_range,
                           show_largest=max_peaks,
                           single_figure=False)
        ev_range = plt.xlim()

    if records_matrix and ax_rec is not None:
        plt.sca(ax_rec)
        context.plot_records_matrix(run_id,
                                    raw=records_matrix == 'raw',
                                    time_range=(events['time'],
                                                events['endtime']),
                                    single_figure=False)
        ax_rec.tick_params(axis='x', rotation=0)
        if not xenon1t:
            # Top vs bottom division
            ax_rec.axhline(straxen.n_top_pmts, c='k')
        if ev_range is not None:
            plt.xlim(*ev_range)

    # Final tweaks
    if ax_s2 is not None:
        ax_s1.tick_params(axis='x', rotation=45)
    if ax_s2 is not None:
        ax_s1.tick_params(axis='x', rotation=45)
    if ax_ev is not None:
        ax_ev.tick_params(axis='x', rotation=0)
    title = (f'Run {run_id}. Time '
             f'{str(events["time"])[:-9]}.{str(events["time"])[-9:]}\n'
             f'{datetime.fromtimestamp(event["time"] / 1e9, tz=pytz.utc)}')
    plt.suptitle(title, y=0.95)
    # NB: reflects panels order
    return (ax_s1, ax_s2, ax_s1_hp_t, ax_s1_hp_b,
            ax_event_info, ax_peak_info, ax_s2_hp_t, ax_s2_hp_b,
            ax_ev,
            ax_rec)


@export
def plot_single_event(context: strax.Context,
                      run_id,
                      events,
                      event_number=None,
                      **kwargs):
    """
    Wrapper for event_display

    :param context: strax.context
    :param run_id: run id
    :param events: dataframe / numpy array of events. Should either be
        length 1 or the event_number argument should be provided
    :param event_number: (optional) int, if provided, only show this
        event number
    :param kwargs: kwargs for events_display
    :return: see events_display
    """
    if event_number is not None:
        events = events[events['event_number'] == event_number]
    if len(events) > 1 or len(events) == 0:
        raise ValueError(f'Make sure to provide an event number or a single '
                         f'event. Got {len(events)} events')

    return context.event_display(run_id,
                                 time_range=(events[0]['time'],
                                             events[0]['endtime']),
                                 **kwargs)


def _scatter_rec(_event,
                 recs=None,
                 scatter_kwargs=None,
                ):
    """Convenient wrapper to show posrec of three algorithms for xenonnt"""
    if recs is None:
        recs = ('mlp', 'cnn', 'gcn')
    elif len(recs) > 5:
        raise ValueError("I only got five markers/colors")
    if scatter_kwargs is None:
        scatter_kwargs = {}
    scatter_kwargs.setdefault('s', 100)
    scatter_kwargs.setdefault('alpha', 0.8)
    shapes = ('v', '^', '>', '<', '*', 'D', "P")
    colors = ('brown', 'orange', 'lightcoral', 'gold', 'lime', 'crimson')
    for _i, _r in enumerate(recs):
        x, y = _event[f's2_x_{_r}'], _event[f's2_y_{_r}']
        if np.isnan(x) or np.isnan(y):
            continue
        plt.scatter(x, y,
                    marker=shapes[_i],
                    c=colors[_i],
                    label=_r.upper(),
                    **scatter_kwargs,
                    )
    plt.legend(loc='best', fontsize="x-small", markerscale=0.5)


# Event display docstrings.
# Let's add them to the corresponding functions

event_docs = """
Make a waveform-display of a given event. Requires events, peaks and
    peaklets (optionally: records). NB: time selection should return
    only one event!

:param context: strax.Context provided by the minianalysis wrapper
:param run_id: run-id of the event
:param events: events, provided by the minianalysis wrapper
:param to_pe: gains, provided by the minianalysis wrapper
:param records_matrix: False (no record matrix), True, or "raw"
    (show raw-record matrix)
:param s2_fuzz: extra time around main S2 [ns]
:param s1_fuzz: extra time around main S1 [ns]
:param max_peaks: max peaks for plotting in the wf plot
:param xenon1t: True: is 1T, False: is nT
:param display_peak_info: tuple, items that will be extracted from
    event and displayed in the event info panel see above for format
:param display_event_info: tuple, items that will be extracted from
    event and displayed in the peak info panel see above for format
:param s1_hp_kwargs: dict, optional kwargs for S1 hitpatterns
:param s2_hp_kwargs: dict, optional kwargs for S2 hitpatterns
:param event_time_limit = overrides x-axis limits of event
    plot
:param plot_all_positions if True, plot best-fit positions
    from all posrec algorithms
"""
event_returns = """
:return: axes used for plotting:
    ax_s1, ax_s2, ax_s1_hp_t, ax_s1_hp_b,
    ax_event_info, ax_peak_info, ax_s2_hp_t, ax_s2_hp_b,
    ax_ev,
    ax_rec
    Where those panels (axes) are:
        - ax_s1, main S1 peak
        - ax_s2, main S2 peak
        - ax_s1_hp_t, S1 top hit pattern
        - ax_s1_hp_b, S1 bottom hit pattern
        - ax_s2_hp_t, S2 top hit pattern
        - ax_s2_hp_b, S2 bottom hit pattern
        - ax_event_info, text info on the event
        - ax_peak_info, text info on the main S1 and S2
        - ax_ev, waveform of the entire event
        - ax_rec, (raw)record matrix (if any otherwise None)
"""

# Add the same docstring to each of these functions
for event_function in (event_display, event_display_simple, _event_display):
    doc = event_function.__doc__
    if doc is not None:
        event_function.__doc__ = doc.format(event_docs=event_docs,
                                            event_returns=event_returns)



import os
import tempfile
from datetime import datetime
from warnings import warn
import pytz
from strax import exporter, to_str_tuple
import gridfs
from tqdm import tqdm
from shutil import move
import hashlib
from pymongo.collection import Collection as pymongo_collection
import utilix
from straxen import uconfig

export, __all__ = exporter()


@export
class GridFsInterface:
    """
    Base class to upload/download the files to a database using GridFS 
    for PyMongo:
    https://pymongo.readthedocs.io/en/stable/api/gridfs/index.html#module-gridfs
    
    This class does the basic shared initiation of the downloader and 
    uploader classes.

    """

    def __init__(self,
                 readonly=True,
                 file_database='files',
                 config_identifier='config_name',
                 collection=None,
                 ):
        """
        GridFsInterface

        :param readonly: bool, can one read or also write to the
            database.
        :param file_database: str, name of the database. Default should
            not be changed.
        :param config_identifier: str, header of the files that are
            saved in Gridfs
        :param collection: pymongo.collection.Collection, (Optional)
            PyMongo DataName Collection to bypass normal initiation
            using utilix. Should be an object of the form:
                pymongo.MongoClient(..).DATABASE_NAME.COLLECTION_NAME
        """
        if collection is None:
            if not readonly:
                # We want admin access to start writing data!
                mongo_url = uconfig.get('rundb_admin', 'mongo_rdb_url')
                mongo_user = uconfig.get('rundb_admin', 'mongo_rdb_username')
                mongo_password = uconfig.get('rundb_admin', 'mongo_rdb_password')
            else:
                # We can safely use the Utilix defaults
                mongo_url = mongo_user = mongo_password = None

            # If no collection arg is passed, it defaults to the 'files'
            # collection, see for more details:
            # https://github.com/XENONnT/utilix/blob/master/utilix/rundb.py
            mongo_kwargs = {
                'url': mongo_url,
                'user': mongo_user,
                'password': mongo_password,
                'database': file_database,
            }
            # We can safely hard-code the collection as that is always
            # the same with GridFS.
            collection = utilix.rundb.xent_collection(
                **mongo_kwargs,
                collection='fs.files')
        else:
            # Check the user input is fine for what we want to do.
            if not isinstance(collection, pymongo_collection):
                raise ValueError('Provide PyMongo collection (see docstring)!')
            assert file_database is None, "Already provided a collection!"

        # Set collection and make sure it can at least do a 'find' operation
        self.collection = collection
        self.test_find()

        # This is the identifier under which we store the files.
        self.config_identifier = config_identifier

        # The GridFS used in this database
        self.grid_fs = gridfs.GridFS(collection.database)

    def get_query_config(self, config):
        """
        Generate identifier to query against. This is just the configs 
        name.

        :param config: str,  name of the file of interest
        :return: dict, that can be used in queries
        """
        return {self.config_identifier: config}

    def document_format(self, config):
        """
        Format of the document to upload

        :param config: str,  name of the file of interest
        :return: dict, that will be used to add the document
        """
        doc = self.get_query_config(config)
        doc.update({
            'added': datetime.now(tz=pytz.utc),
        })
        return doc

    def config_exists(self, config):
        """
        Quick check if this config is already saved in the collection

        :param config: str,  name of the file of interest
        :return: bool, is this config name stored in the database
        """
        query = self.get_query_config(config)
        return self.collection.count_documents(query) > 0

    def md5_stored(self, abs_path):
        """
        NB: RAM intensive operation!
        Carefully compare if the MD5 identifier is the same as the file 
        as stored under abs_path.

        :param abs_path: str, absolute path to the file name
        :return: bool, returns if the exact same file is already stored
            in the database

        """
        if not os.path.exists(abs_path):
            # A file that does not exist does not have the same MD5
            return False
        query = {'md5': self.compute_md5(abs_path)}
        return self.collection.count_documents(query) > 0

    def test_find(self):
        """
        Test the connection to the self.collection to see if we can 
        perform a collection.find operation.
        """
        if self.collection.find_one(projection="_id") is None:
            raise ConnectionError('Could not find any data in this collection')

    def list_files(self):
        """
        Get a complete list of files that are stored in the database

        :return: list, list of the names of the items stored in this
            database

        """
        return [doc[self.config_identifier]
                for doc in
                self.collection.find(
                    projection=
                    {self.config_identifier: 1})
                if self.config_identifier in doc
                ]

    @staticmethod
    def compute_md5(abs_path):
        """
        NB: RAM intensive operation!
        Get the md5 hash of a file stored under abs_path

        :param abs_path: str, absolute path to a file
        :return: str, the md5-hash of the requested file
        """
        # This function is copied from:
        # stackoverflow.com/questions/3431825/generating-an-md5-checksum-of-a-file

        if not os.path.exists(abs_path):
            # if there is no file, there is nothing to compute
            return ""
        # Also, disable all the  Use of insecure MD2, MD4, MD5, or SHA1
        # hash function violations in this function.
        # bandit: disable=B303
        hash_md5 = hashlib.md5()
        with open(abs_path, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()


@export
class MongoUploader(GridFsInterface):
    """
    Class to upload files to GridFs
    """

    def __init__(self, readonly=False, *args, **kwargs):
        # Same as parent. Just check the readonly_argument
        if readonly:
            raise PermissionError(
                "How can you upload if you want to operate in readonly?")
        super().__init__(*args, readonly=readonly, **kwargs)

    def upload_from_dict(self, file_path_dict):
        """
        Upload all files in the dictionary to the database.

        :param file_path_dict: dict, dictionary of paths to upload. The 
            dict should be of the format:
            file_path_dict = {'config_name':  '/the_config_path', ...}

        :return: None
        """
        if not isinstance(file_path_dict, dict):
            raise ValueError(f'file_path_dict must be dict of form '
                             f'"dict(NAME=ABSOLUTE_PATH,...)". Got '
                             f'{type(file_path_dict)} instead')

        for config, abs_path in tqdm(file_path_dict.items()):
            # We need to do this expensive check here. It is not enough
            # to just check that the file is stored under the 
            # 'config_identifier'. What if the file changed? Then we 
            # want to upload a new file! Otherwise we could have done 
            # the self.config_exists-query. If it turns out we have the 
            # exact same file, forget about uploading it.
            if self.config_exists(config) and self.md5_stored(abs_path):
                continue
            else:
                # This means we are going to upload the file because its
                # not stored yet.
                try:
                    self.upload_single(config, abs_path)
                except (CouldNotLoadError, ConfigTooLargeError):
                    # Perhaps we should fail then?
                    warn(f'Cannot upload {config}')

    def upload_single(self, config, abs_path):
        """
        Upload a single file to gridfs

        :param config: str, the name under which this file should be
            stored

        :param abs_path: str, the absolute path of the file 
        """
        doc = self.document_format(config)
        if not os.path.exists(abs_path):
            raise CouldNotLoadError(f'{abs_path} does not exits')

        print(f'uploading {config}')
        with open(abs_path, 'rb') as file:
            self.grid_fs.put(file, **doc)


@export
class MongoDownloader(GridFsInterface):
    """
    Class to download files from GridFs
    """

    def __init__(self,
                 store_files_at=None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        # We are going to set a place where to store the files. It's 
        # either specified by the user or we use these defaults:
        if store_files_at is None:
            store_files_at = ('/tmp/straxen_resource_cache/',
                              '/dali/lgrandi/strax/resource_cache',
                              './resource_cache',  
                              )
        elif not isinstance(store_files_at, (tuple, str, list)):
            raise ValueError(f'{store_files_at} should be tuple of paths!')
        elif isinstance(store_files_at, str):
            store_files_at = to_str_tuple(store_files_at)

        self.storage_options = store_files_at

    def download_single(self,
                        config_name: str,
                        human_readable_file_name=False):
        """
        Download the config_name if it exists

        :param config_name: str, the name under which the file is stored

        :param human_readable_file_name: bool, store the file also under
            it's human readable name. It is better not to use this as
            the user might not know if the version of the file is the
            latest.

        :return: str, the absolute path of the file requested
        """
        if self.config_exists(config_name):
            # Query by name
            query = self.get_query_config(config_name)
            try:
                # This could return multiple since we upload files if 
                # they have changed again! Therefore just take the last.
                fs_object = self.grid_fs.get_last_version(**query)
            except gridfs.NoFile as e:
                raise CouldNotLoadError(
                    f'{config_name} cannot be downloaded from GridFs') from e

            # Ok, so we can open it. We will store the file under it's 
            # md5-hash as that allows to easily compare if we already 
            # have the correct file.
            if human_readable_file_name:
                target_file_name = config_name
            else:
                target_file_name = fs_object.md5

            for cache_folder in self.storage_options:
                possible_path = os.path.join(cache_folder, target_file_name)
                if os.path.exists(possible_path):
                    # Great! This already exists. Let's just return
                    # where it is stored.
                    return possible_path

            # Apparently the file does not exist, let's find a place to
            # store the file and download it.
            store_files_at = self._check_store_files_at(self.storage_options)
            destination_path = os.path.join(store_files_at, target_file_name)

            # Let's open a temporary directory, download the file, and
            # try moving it to the destination_path. This prevents
            # simultaneous writes of the same file.
            with tempfile.TemporaryDirectory() as temp_directory_name:
                temp_path = os.path.join(temp_directory_name, target_file_name)

                with open(temp_path, 'wb') as stored_file:
                    # This is were we do the actual downloading!
                    warn(f'Downloading {config_name} to {destination_path}')
                    stored_file.write(fs_object.read())

                if not os.path.exists(destination_path):
                    # Move the file to the place we want to store it.
                    move(temp_path, destination_path)
            return destination_path

        else:
            raise ValueError(f'Config {config_name} cannot be downloaded '
                             f'since it is not stored')

    def get_abs_path(self, config_name):
        return self.download_single(config_name)

    def download_all(self):
        """Download all the files that are stored in the mongo collection"""
        raise NotImplementedError('This feature is disabled for now')
        # Disable the inspection of `Unreachable code`
        # pylint: disable=unreachable
        for config in self.list_files():
            self.download_single(config)

    @staticmethod
    def _check_store_files_at(cache_folder_alternatives):
        """
        Iterate over the options in cache_options until we find a folder
            where we can store data. Order does matter as we iterate
            until we find one folder that is willing.

        :param cache_folder_alternatives: tuple, this tuple must be a
            list of paths one can try to store the downloaded data

        :return: str, the folder that we can write to.
        """
        if not isinstance(cache_folder_alternatives, (tuple, list)):
            raise ValueError('cache_folder_alternatives must be tuple')
        for folder in cache_folder_alternatives:
            if not os.path.exists(folder):
                try:
                    os.makedirs(folder)
                except (PermissionError, OSError):
                    continue
            if os.access(folder, os.W_OK):
                return folder
        raise PermissionError(
            f'Cannot write to any of the cache_folder_alternatives: '
            f'{cache_folder_alternatives}')


class CouldNotLoadError(Exception):
    """Raise if we cannot load this kind of data"""
    # Disable the inspection of 'Unnecessary pass statement'
    # pylint: disable=unnecessary-pass
    pass


class ConfigTooLargeError(Exception):
    """Raise if the data is to large to be uploaded into mongo"""
    # Disable the inspection of 'Unnecessary pass statement'
    # pylint: disable=unnecessary-pass
    pass
